package org.implab.gradle.containers.tasks;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.ExecutionException;

import org.gradle.api.Action;
import org.gradle.api.file.RegularFileProperty;
import org.gradle.api.provider.ListProperty;
import org.gradle.api.provider.Property;
import org.gradle.api.tasks.Internal;
import org.gradle.api.tasks.TaskAction;
import org.implab.gradle.containers.cli.RedirectFrom;
import org.implab.gradle.containers.cli.RedirectTo;
import org.implab.gradle.containers.dsl.OptionsMixin;
import org.implab.gradle.containers.dsl.RedirectFromSpec;
import org.implab.gradle.containers.dsl.RedirectToSpec;
import org.implab.gradle.containers.dsl.VolumeSpec;

public abstract class RunContainer extends DockerCliTask implements OptionsMixin {

    private final RedirectToSpec redirectStderr = new RedirectToSpec();

    private final RedirectToSpec redirectStdout = new RedirectToSpec();

    private final RedirectFromSpec redirectStdin = new RedirectFromSpec();

    private boolean transientContainer = true;

    private boolean detached = false;

    private boolean interactive = false;

    @Internal
    public abstract Property<Object> getImageName();

    @Internal
    public abstract ListProperty<String> getCommandLine();

    @Internal
    public RedirectFromSpec getStdin() {
        return redirectStdin;
    }

    @Internal
    public RedirectToSpec getStdout() {
        return redirectStdout;
    }

    @Internal
    public RedirectToSpec getStderr() {
        return redirectStderr;
    }

    @Internal
    public boolean isTransientContainer() {
        return transientContainer;
    }

    public void setTransientContainer(boolean value) {
        transientContainer = value;
    }

    @Internal
    public boolean isDetached() {
        // if IO was redirected the container should run in foreground
        if (redirectStdin.isRedirected() || redirectStdout.isRedirected() || redirectStderr.isRedirected())
            return false;

        return detached;
    }

    public void setDetached(boolean value) {
        detached = value;
    }

    @Internal
    public boolean isInteractive() {
        // enable interactive mode when processing standard input
        return redirectStdin.isRedirected() || interactive;
    }

    @Internal
    public abstract RegularFileProperty getContainerIdFile();

    public RunContainer() {
        getContainerIdFile().convention(() -> new File(getTemporaryDir(), "cid"));
    }

    public void volume(Action<VolumeSpec> configure) {
        getOptions().add("-v");
        getOptions().add(provider(() -> {
            var volumeSpec = getObjectFactory().newInstance(VolumeSpec.class);
            configure.execute(volumeSpec);
            return volumeSpec.resolveSpec();
        }));
    }

    void commandLine(String... args) {
        getCommandLine().addAll(args);
    }

    @Override
    protected Optional<RedirectTo> getStdoutRedirection() {
        return redirectStdout.getRedirection().or(super::getStdoutRedirection);
    }

    @Override
    protected Optional<RedirectTo> getStderrRedirection() {
        return redirectStderr.getRedirection().or(super::getStderrRedirection);
    }

    @Override
    protected Optional<RedirectFrom> getStdinRedirection() {
        return redirectStdin.getRedirection();
    }

    @TaskAction
    public void run() throws InterruptedException, IOException, ExecutionException {
        var params = new ArrayList<>(getOptions().get());

        if (isInteractive())
            params.add("--interactive");
        if (isTransientContainer())
            params.add("--rm");
        if (isDetached())
            params.add("--detach");
        if (getContainerIdFile().isPresent()) {
            getContainerIdFile().getAsFile().get().delete();

            params.addAll(List.of("--cidfile", getContainerIdFile().get().getAsFile().toString()));
        }

        var spec = docker().runImage(
                getImageName().map(Object::toString).get(),
                params,
                getCommandLine().get());

        // if process will run in foreground, then setup default redirects
        redirectStdout.getRedirection()
                .or(this::loggerInfoRedirect)
                .ifPresent(spec::redirectStdout);

        redirectStderr.getRedirection()
                .or(this::loggerErrorRedirect)
                .ifPresent(spec::redirectStderr);

        redirectStdin.getRedirection().ifPresent(spec::redirectStdin);

        getLogger().info("Staring: {}", spec.command());

        // runs the command and checks the error code
        spec.exec();
    }

}
